from crosswalk.exceptions import NestedAttributesError, ReservedKeyError


def validate_shallow_dict(value):
    """Only allow shallow dictionaries."""
    if isinstance(value, dict):
        for key in value:
            if isinstance(value[key], dict):
                raise NestedAttributesError(
                    "Nested attributes are not allowed."
                )


def validate_no_reserved_keys(value):
    reserved_keys = [
        'alias_for',
        'aliased',
        'attributes',
        'created',
        'domain',
        'entity',
        'match_score',
        'superseded_by',
        'uuid',
    ]

    if isinstance(value, dict):
        for key in value:
            if key in reserved_keys:
                raise ReservedKeyError(
                    "Reserved key in attributes."
                )


def full_validation(value):
    validate_shallow_dict(value)
    validate_no_reserved_keys(value)



from crosswalk.models import ApiUser
from django.contrib import admin


class ApiUserAdmin(admin.ModelAdmin):
    fields = ('user',)
    list_display = ('user', 'token',)


admin.site.register(ApiUser, ApiUserAdmin)



from crosswalk.authentication import AuthenticatedView
from crosswalk.exceptions import NestedAttributesError, ReservedKeyError
from crosswalk.models import Domain, Entity
from crosswalk.serializers import EntitySerializer
from crosswalk.validators import (validate_no_reserved_keys,
                                  validate_shallow_dict)
from django.core.exceptions import ObjectDoesNotExist
from rest_framework import status
from rest_framework.response import Response


class BulkCreate(AuthenticatedView):
    def post(self, request, domain):
        try:
            domain = Domain.objects.get(slug=domain)
        except ObjectDoesNotExist:
            return Response(
                "Domain not found.",
                status=status.HTTP_200_OK
            )

        entities = request.data.copy()

        entity_objects = []

        for entity in entities:
            uuid = entity.pop("uuid", None)

            # Validate entity attributes before creating in bulk
            try:
                validate_shallow_dict(entity)
            except NestedAttributesError:
                return Response(
                    "Cannot create entity with nested attributes.",
                    status=status.HTTP_400_BAD_REQUEST
                )
            try:
                validate_no_reserved_keys(entity)
            except ReservedKeyError:
                return Response(
                    "Reserved key found in entity attributes.",
                    status=status.HTTP_400_BAD_REQUEST
                )

            entity_objects.append(
                Entity(
                    uuid=uuid,
                    domain=domain,
                    attributes=entity,
                    created_by=request.user
                )
            )

        created_entities = Entity.objects.bulk_create(entity_objects)

        return Response({
            "entities": [
                {
                    "entity": EntitySerializer(entity).data,
                    "created": True
                }
                for entity in created_entities
            ]
        }, status=status.HTTP_200_OK)



from django.urls import include, path
from rest_framework import routers

from .views import (AliasOrCreate, BestMatch, BestMatchOrCreate, BulkCreate,
                    ClientCheck, DeleteMatch, UpdateMatch)
from .viewsets import DomainViewSet, EntityDomainViewSet, EntityViewSet

router = routers.DefaultRouter()

router.register('domains', DomainViewSet)

entity_detail = EntityViewSet.as_view({
    'get': 'retrieve',
    'put': 'update',
    'patch': 'partial_update',
    'delete': 'destroy'
})
entity_domain_list = EntityDomainViewSet.as_view({
    'get': 'list',
    'post': 'create'
})
entity_domain_detail = EntityDomainViewSet.as_view({
    'get': 'retrieve',
    'put': 'update',
    'patch': 'partial_update',
    'delete': 'destroy'
})

urlpatterns = [
    path('api/', include(router.urls)),
    path(
        'api/domains/<slug:domain>/entities/', entity_domain_list,
        name="crosswalk-entity-domain-lists"
    ),
    path(
        'api/domains/<slug:domain>/entities/bulk-create/',
        BulkCreate.as_view()
    ),
    path(
        'api/domains/<slug:domain>/entities/best-match/',
        BestMatch.as_view()
    ),
    path(
        'api/domains/<slug:domain>/entities/best-match-or-create/',
        BestMatchOrCreate.as_view()
    ),
    path(
        'api/domains/<slug:domain>/entities/delete-match/',
        DeleteMatch.as_view()
    ),
    path(
        'api/domains/<slug:domain>/entities/update-match/',
        UpdateMatch.as_view()
    ),
    path(
        'api/domains/<slug:domain>/entities/alias-or-create/',
        AliasOrCreate.as_view()
    ),
    path(
        'api/domains/<slug:domain>/entities/<uuid:pk>/', entity_domain_detail,
        name="crosswalk-entity-domain-detail"
    ),
    path(
        'api/entities/<uuid:pk>/', entity_detail,
        name="crosswalk-entity-detail"
    ),
    path('api/client-check/', ClientCheck.as_view()),
]



from rest_framework import viewsets
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response

from crosswalk.authentication import TokenAuthentication

from .models import Domain, Entity
from .serializers import DomainSerializer, EntitySerializer


class AuthenticatedViewSet(viewsets.ModelViewSet):
    permission_classes = (IsAuthenticated,)
    authentication_classes = (TokenAuthentication,)
    paginator = None


class DomainViewSet(AuthenticatedViewSet):
    serializer_class = DomainSerializer
    queryset = Domain.objects.all()
    lookup_field = "slug"


class EntityDomainViewSet(AuthenticatedViewSet):
    serializer_class = EntitySerializer

    def get_queryset(self):
        domain = self.kwargs["domain"]
        return Entity.objects.filter(domain__name=domain)

    def list(self, request, domain):
        """Allow passing block attr params to filter queryset."""
        params = request.query_params.copy()
        queryset = self.get_queryset().filter(attributes__contains=params)
        serializer = self.serializer_class(queryset, many=True)
        return Response(serializer.data)


class EntityViewSet(AuthenticatedViewSet):
    serializer_class = EntitySerializer
    queryset = Entity.objects.all()

    def partial_update(self, request, pk=None):
        """
        Merge attributes dict with existing attributes.
        """
        instance = self.queryset.get(pk=pk)
        new_attributes = {
            **instance.attributes,
            **request.data.get("attributes", {}),
        }
        serializer = self.serializer_class(
            instance,
            data={
                "attributes": new_attributes,
                "alias_for": request.data.get("alias_for", instance.alias_for),
                "superseded_by": request.data.get(
                    "superseded_by", instance.superseded_by
                ),
            },
            partial=True,
        )
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response(serializer.data)
